FIELD_IS_REQUIRED = "Field is required."
ANY_FIELD_IS_REQUIRED = "One of these fields must be provided."
DATE_IS_NOT_AFTER = "Date is too late."
DATE_IS_NOT_BEFORE = "Date is too early."
FIELD_MUST_BE_BLANK = "Field must be blank."
